"""Main interface to the physical constants store."""

from dataclasses import make_dataclass
import json
from pathlib import Path

from iris.coord_systems import GeogCS
from iris.cube import Cube
import iris.fileformats
import numpy as np

from ..exceptions import ArgumentError, LoadError, _warn

__all__ = ("add_planet_conf_to_cubes", "get_planet_radius", "init_const")

CONST_DIR = Path(__file__).parent / "store"

DERIVED_CONST = {
    "dry_air_gas_constant": (
        lambda slf: slf.molar_gas_constant / slf.dry_air_molecular_weight,
        "J kg-1 K-1",
    ),
    "molecular_weight_ratio": (
        lambda slf: slf.condensible_molecular_weight
        / slf.dry_air_molecular_weight,
        "1",
    ),
    "kappa": (
        lambda slf: slf.dry_air_gas_constant / slf.dry_air_spec_heat_press,
        "1",
    ),  # poisson_exponent
    "planet_rotation_rate": (
        lambda slf: (slf.day / (2 * np.pi)) ** (-1),
        "s-1",
    ),
}


class ConstContainer:
    """Base class for creating dataclasses and storing planetary constants."""

    def __repr__(self):
        """Create custom repr."""
        cubes_str = ", ".join(
            [
                (
                    f"{getattr(self, _field).long_name}"
                    f" [{getattr(self, _field).units}]"
                )
                for _field in self.__dataclass_fields__
            ]
        )
        return f"{self.__class__.__name__}({cubes_str})"

    def __post_init__(self):
        """Do things automatically after __init__()."""
        self._convert_to_iris_cubes()
        self._derive_const()

    def _convert_to_iris_cubes(self):
        """Loop through fields and convert each of them to `iris.cube.Cube`."""
        for name in self.__dataclass_fields__:
            _field = getattr(self, name)
            cube = Cube(
                data=_field.get("value"),
                units=_field.get("units", 1),
                long_name=name,
            )
            object.__setattr__(self, name, cube)

    def _derive_const(self):
        """Not fully implemented yet."""
        for name, recipe in DERIVED_CONST.items():
            func, units = recipe
            try:
                cube = func(self)
                cube.convert_units(units)
                cube.rename(name)
                object.__setattr__(self, name, cube)
            except AttributeError:
                pass


def _read_const_file(name, directory=CONST_DIR):
    """Read constants from the JSON file."""
    if not isinstance(directory, Path):
        raise ArgumentError("directory must be a pathlib.Path object")
    try:
        with (directory / name).with_suffix(".json").open("r") as fp:
            list_of_dicts = json.load(fp)
        # transform the list of dictionaries into a dictionary
        const_dict = {}
        for vardict in list_of_dicts:
            const_dict[vardict["name"]] = {
                k: v for k, v in vardict.items() if k != "name"
            }
        return const_dict
    except FileNotFoundError:
        raise LoadError(
            f"JSON file for {name} configuration not found,"
            f"check the directory: {directory}"
        )


def init_const(name="general", directory=None):
    """
    Create a dataclass with a given set of constants.

    Parameters
    ----------
    name: str, optional
        Name of the constants set.
        Should be identical to the JSON file name (w/o the .json extension).
        If not given, only general physical constants are returned.
    directory: pathlib.Path, optional
        Path to a folder with JSON files with constants for a specific planet.

    Returns
    -------
    Dataclass with constants as iris cubes.

    Examples
    --------
    >>> c = init_const('earth')
    >>> c
    EarthConstants(gravity [m s-2], radius [m], day [s], ...)
    >>> c.gravity
    <iris 'Cube' of gravity / (m s-2) (scalar cube)>
    """
    cls_name = f"{name.capitalize()}Constants"
    if directory is None:
        # use default directory
        kw = {}
    else:
        kw = {"directory": directory}
    # transform the list of dictionaries into a dictionary
    const_dict = _read_const_file("general")  # TODO: make this more flexible?
    if name != "general":
        const_dict.update(_read_const_file(name, **kw))
    kls = make_dataclass(
        cls_name,
        fields=[*const_dict.keys()],
        bases=(ConstContainer,),
        frozen=True,
        repr=False,
    )
    return kls(**const_dict)


def get_planet_radius(cube, default=iris.fileformats.pp.EARTH_RADIUS):
    """Get the planet radius in metres from cube attributes."""
    cs = cube.coord_system("CoordSystem")
    if cs is not None:
        r = cs.semi_major_axis
    else:
        try:
            r = cube.attributes["planet_conf"].radius.copy()
            r.convert_units("m")
            r = float(r.data)
        except (KeyError, LoadError):
            _warn("Using default radius")
            r = default
    return r


def add_planet_conf_to_cubes(cubelist, const):
    """
    Add constants container to the cube attributes and adjust its coord system.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        List of cubes containing a cube of zonal velocity (u).
    const: aeolus.const.const.ConstContainer, optional
        Constainer with the relevant planetary constants.
    """
    const.radius.convert_units("m")
    _coord_system = GeogCS(semi_major_axis=const.radius.data)
    for cube in cubelist:
        # add constants to cube attributes
        cube.attributes["planet_conf"] = const
        for coord in cube.coords():
            if coord.coord_system:
                # Replace coordinate system with the radius from `self.const`
                coord.coord_system = _coord_system



"""Prepare data in spherical coordinates for pyvista plotting."""

from typing import Optional

import geovista as gv
import iris
import numpy as np
import pyvista as pv
from pyvista import grid_from_sph_coords, transform_vectors_sph_to_cart

from ..coord import _cell_bounds
from ..exceptions import _warn
from ..model import um
from ..model.base import Model

__all__ = (
    "grid_for_scalar_cube_sph",
    "grid_for_vector_cubes_sph",
    "ugrid_mesh_to_gv_mesh",
)


def grid_for_scalar_cube_sph(
    cube: iris.cube.Cube,
    z_scale: Optional[float] = 1,
    z_offset: Optional[float] = 0,
    grid: Optional[pv.StructuredGrid] = None,
    label: Optional[str] = "scalar3d",
    model: Optional[Model] = um,
) -> pv.PolyData:
    """
    Create a `pyvista` grid for an `iris` cube in spherical coordinates.

    Parameters
    ----------
    cube: iris.cube.Cube
        2D or 3D cube with (longitude, latitude, [level_height]) coordinates.
    z_scale: float, optional
        Scaling factor for the vertical level dimension.
    z_offset: float, optional
        Scaling offset for the vertical level dimension.
    grid: pyvista.StructuredGrid, optional
        If given, add data to the existing grid, otherwise create a new one
        from the input cube's coordinates.
    label: str, optional
        Label for the data within the grid.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    pyvista.StructuredGrid
       PyVista grid with data in the cell_data attribute.
    """
    if grid is None:
        lons = _cell_bounds(cube.coord(model.x).points)
        lats = 90.0 - _cell_bounds(cube.coord(model.y).points)
        if cube.ndim == 3:
            # TODO check if _cell_bounds should be here
            levels = _cell_bounds(cube.coord(model.z).points)
        else:
            levels = np.array([0])
        levels = z_scale * levels + z_offset

        grid = grid_from_sph_coords(lons, lats, levels)

    # Add data arras to grid
    if label in grid.cell_data:
        _warn(f"Label '{label}' exists in {grid}")
    grid.cell_data[label] = np.array(cube.data).swapaxes(-2, -1).ravel("C")
    return grid


def grid_for_vector_cubes_sph(
    u: iris.cube.Cube,
    v: iris.cube.Cube,
    w: iris.cube.Cube,
    vector_scale: Optional[float] = 1,
    vertical_wind_scale: Optional[float] = 1,
    z_scale: Optional[float] = 1,
    z_offset: Optional[float] = 0,
    xstride: Optional[int] = 1,
    ystride: Optional[int] = 1,
    grid: Optional[pv.StructuredGrid] = None,
    label: Optional[str] = "vector3d",
    model: Optional[Model] = um,
) -> pv.PolyData:
    """
    Take winds in spherical coordinates and create a `pyvista` grid for them.

    Parameters
    ----------
    u: iris.cube.Cube
        2D or 3D cube of x-wind component (zonal wind).
    v: iris.cube.Cube
        2D or 3D cube of y-wind component (meridional wind).
    w: iris.cube.Cube
        2D or 3D cube of z-wind component (vertical wind).
    vector_scale: float, optional
        Scaling factor for vectors.
    vertical_wind_scale: float, optional
        Scaling factor for the vertical wind component (for better visibility).
    z_scale: float, optional
        Scaling factor for the vertical level dimension.
    z_offset: float, optional
        Scaling offset for the vertical level dimension.
    xstride: float, optional
        Stride along the longitude axis.
    ystride: float, optional
        Stride along the latitude axis.
    grid: pyvista.StructuredGrid, optional
        If given, add data to the existing grid, otherwise create a new one
        from the input cube's coordinates.
    label: str, optional
        Label for the data within the grid.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    pyvista.StructuredGrid
       PyVista grid with vector data in the point_data attribute.
    """
    assert (u.shape == v.shape) and (
        u.shape == w.shape
    ), "Wind components should have the same array size!"

    lons = u.coord(model.x).points
    lats = 90.0 - u.coord(model.y).points
    levels = z_scale * u.coord(model.z).points + z_offset

    # Stride vectors
    lons_s = lons[::xstride]
    lats_s = lats[::ystride]

    inv_axes = [*range(u.ndim - 1, -1, -1)]

    u_sdata = u.data[..., ::ystride, ::xstride].transpose(inv_axes)
    v_sdata = v.data[..., ::ystride, ::xstride].transpose(inv_axes)
    w_sdata = w.data[..., ::ystride, ::xstride].transpose(inv_axes)

    # Rescale vertical wind for better viz
    w_sdata *= vertical_wind_scale

    # Create a stack of vectors transformed from spherical to cartesian coords
    vectors = np.stack(
        [
            i.transpose(inv_axes).swapaxes(u.ndim - 2, u.ndim - 1).ravel("C")
            for i in transform_vectors_sph_to_cart(
                lons_s, lats_s, levels, u_sdata, -v_sdata, w_sdata
            )
        ],
        axis=1,
    )
    # Scale vectors
    vectors *= vector_scale

    if grid is None:
        grid = grid_from_sph_coords(lons_s, lats_s, levels)

    # Add vectors to the grid
    grid.point_data[label] = vectors
    return grid


def ugrid_mesh_to_gv_mesh(
    topology: iris.experimental.ugrid.mesh.Mesh, **kwargs
) -> pv.PolyData:
    """Create a geovista mesh from the LFRic mesh."""
    face_node = topology.face_node_connectivity

    indices = face_node.indices_by_location()

    lons, lats = topology.node_coords

    return gv.Transform.from_unstructured(
        lons.points,
        lats.points,
        indices,
        start_index=face_node.start_index,
        **kwargs,
    )



"""Exceptions specific to aeolus package."""

from warnings import warn


def _warn(msg):
    """Raise an aeolus-related warning."""
    warn(msg, AeolusWarning, stacklevel=2)


class AeolusWarning(UserWarning):
    """Base class for warnings in aeolus package."""

    pass


class DeprecatedWarning(AeolusWarning):
    """Warning for a deprecated feature."""

    pass


class AeolusError(Exception):
    """Base class for errors in aeolus package."""

    pass


class NotYetImplementedError(AeolusError):
    """
    Raised by missing functionality.

    Different meaning to NotImplementedError, which is for abstract methods.
    """

    pass


class ArgumentError(AeolusError):
    """Raised when argument type is not recognised or is not allowed."""

    pass


class LoadError(AeolusError):
    """Raised when input files or directories are not found."""

    pass


class BoundaryError(AeolusError):
    """Raised when there is an error with geographical regions."""

    pass


class NotFoundError(AeolusError):
    """Raised when metadata is not found."""

    pass


class UnitFormatError(AeolusError):
    """Raised when cube units cannot be formatted."""

    pass


class MissingCubeError(AeolusError):
    """Raised when cubes required for a calculation are missing."""

    pass


class BadCoordinateError(AeolusError):
    """Raised when coordinates are inconsistent."""

    pass



"""Some commonly used diagnostics in atmospheric science."""

from cf_units import Unit
from iris.analysis.calculus import _coord_cos
from iris.analysis.maths import add, apply_ufunc, multiply
from iris.exceptions import ConstraintMismatchError as ConMisErr
from iris.util import reverse
import numpy as np

from ..const import init_const
from ..coord import coord_to_cube, ensure_bounds, regrid_3d
from ..exceptions import ArgumentError, MissingCubeError
from ..meta import const_from_attrs, preserve_shape, update_metadata
from ..model import um
from ..subset import _dim_constr
from .calculus import d_dz, integrate
from .stats import cumsum, spatial_mean, time_mean, zonal_mean

__all__ = (
    "air_density",
    "air_potential_temperature",
    "air_pressure",
    "air_temperature",
    "bond_albedo",
    "calc_derived_cubes",
    "dry_lapse_rate",
    "flux",
    "geopotential_height",
    "ghe_norm",
    "greenhouse_effect",
    "heat_redist_eff",
    "horiz_wind_cmpnts",
    "meridional_mass_streamfunction",
    "precip_sum",
    "sfc_net_energy",
    "sfc_water_balance",
    "sigma_p",
    "toa_cloud_radiative_effect",
    "toa_eff_temp",
    "toa_net_energy",
    "water_path",
    "wind_rot_div",
    "wind_speed",
    "zonal_mass_streamfunction",
)


def _precip_name_mapping(model=um):
    """Generate lists of variable names for `precip_sum()`."""
    return {
        "total": [model.ls_rain, model.ls_snow, model.cv_rain, model.cv_snow],
        "conv": [model.cv_rain, model.cv_snow],
        "stra": [model.ls_rain, model.ls_snow],
        "rain": [model.ls_rain, model.cv_rain],
        "snow": [model.ls_snow, model.cv_snow],
    }


@update_metadata(name="cos_lat", units="1")
def lat_cos(cube, model=um):
    """Convert the latitude coordinate to a cube and apply cosine to it."""
    lat_cube = coord_to_cube(cube, model.y, broadcast=False)
    lat_cos_cube = apply_ufunc(np.cos, apply_ufunc(np.deg2rad, lat_cube))
    return lat_cos_cube


@update_metadata(units="K")
def air_temperature(cubelist, const=None, model=um):
    """
    Get the real temperature from the given cube list.

    If not present, it is attempted to calculate it from other variables,
    Exner pressure or air pressure, and potential temperature.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes containing temperature.
    const: aeolus.const.const.ConstContainer, optional
        Must have `reference_surface_pressure` and `dry_air_gas_constant`.
        If not given, an attempt is made to retrieve it from cube attributes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of air temperature.
    """
    try:
        return cubelist.extract_cube(model.temp)
    except ConMisErr:
        try:
            thta = cubelist.extract_cube(model.thta)
        except ConMisErr:
            raise MissingCubeError(
                f"Unable to get air temperature from {cubelist}"
            )

        if len(cubelist.extract(model.exner)) == 1:
            exner = cubelist.extract_cube(model.exner)
        elif len(cubelist.extract(model.pres)) == 1:
            if const is None:
                const = thta.attributes["planet_conf"]
            pres = cubelist.extract_cube(model.pres)
            exner = (pres / const.reference_surface_pressure) ** (
                const.dry_air_gas_constant / const.dry_air_spec_heat_press
            ).data
        else:
            raise MissingCubeError(
                f"Unable to get air temperature from {cubelist}"
            )
        temp = thta * exner
        temp.rename(model.temp)
        return temp


@update_metadata(units="K")
def air_potential_temperature(cubelist, const=None, model=um):
    """
    Get the potential temperature from the given cube list.

    If not present, it is attempted to calculate it from other variables,
    Exner pressure or air pressure, and real temperature.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes containing temperature.
    const: aeolus.const.const.ConstContainer, optional
        Must have `reference_surface_pressure` and `dry_air_gas_constant`.
        If not given, an attempt is made to retrieve it from cube attributes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of air potential temperature.
    """
    try:
        return cubelist.extract_cube(model.thta)
    except ConMisErr:
        try:
            temp = cubelist.extract_cube(model.temp)
        except ConMisErr:
            raise MissingCubeError(
                f"Unable to get air potential temperature from {cubelist}"
            )

        if len(cubelist.extract(model.exner)) == 1:
            exner = cubelist.extract_cube(model.exner)
        elif len(cubelist.extract(model.pres)) == 1:
            if const is None:
                const = temp.attributes["planet_conf"]
            pres = cubelist.extract_cube(model.pres)
            exner = (pres / const.reference_surface_pressure) ** (
                const.dry_air_gas_constant / const.dry_air_spec_heat_press
            ).data
        else:
            raise MissingCubeError(
                f"Unable to get air potential temperature from {cubelist}"
            )
        thta = temp / exner
        thta.rename(model.thta)
        thta.convert_units("K")
        return thta


@const_from_attrs()
@update_metadata(units="Pa")
def air_pressure(cubelist, const=None, model=um):
    """
    Get pressure from the given cube list.

    If not present, it is attempted to calculate it from other variables,
    such as Exner pressure or air potential temperature and real temperature.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes containing related variables.
    const: aeolus.const.const.ConstContainer, optional
        Must have `reference_surface_pressure` and `dry_air_gas_constant`.
        If not given, an attempt is made to retrieve it from cube attributes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of air pressure.
    """
    try:
        return cubelist.extract_cube(model.pres)
    except ConMisErr:
        try:
            exner = cubelist.extract_cube(model.exner)
            pres = (
                const.reference_surface_pressure
                * exner
                ** (
                    const.dry_air_spec_heat_press / const.dry_air_gas_constant
                ).data
            )
        except ConMisErr:
            try:
                temp = cubelist.extract_cube(model.temp)
                thta = cubelist.extract_cube(model.temp)
                exner = temp / thta
                pres = (
                    const.reference_surface_pressure
                    * exner
                    ** (
                        const.dry_air_spec_heat_press
                        / const.dry_air_gas_constant
                    ).data
                )
            except ConMisErr:
                raise MissingCubeError(
                    f"Unable to calculate air pressure from {cubelist}"
                )
        pres.rename(model.pres)
        return pres


@update_metadata(units="kg m-3")
def air_density(cubelist, const=None, model=um):
    """
    Get air density from the given cube list.

    If not present, try to calculate it from pressure and temperature.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes containing temperature.
    const: aeolus.const.const.ConstContainer, optional
        Must have `dry_air_gas_constant` as an attribute.
        If not given, an attempt is made to retrieve it from cube attributes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of air density.
    """
    try:
        return cubelist.extract_cube(model.dens)
    except ConMisErr:
        try:
            temp = cubelist.extract_cube(model.temp)
            pres = cubelist.extract_cube(model.pres)
            if const is None:
                const = pres.attributes["planet_conf"]
            rho = pres / (const.dry_air_gas_constant * temp)
            rho.rename(model.dens)
            return rho
        except ConMisErr:
            _msg = (
                f"Unable to get variables from\n{cubelist}"
                "\nto calculate air density"
            )
            raise MissingCubeError(_msg)


@const_from_attrs()
def calc_derived_cubes(cubelist, const=None, model=um):
    """Calculate additional variables."""
    try:
        cubelist.extract_cube(model.temp)
    except ConMisErr:
        cubelist.append(air_temperature(cubelist, const=const, model=model))
    try:
        cubelist.extract_cube(model.thta)
    except ConMisErr:
        cubelist.append(
            air_potential_temperature(cubelist, const=const, model=model)
        )
    try:
        cubelist.extract_cube(model.pres)
    except ConMisErr:
        cubelist.append(air_pressure(cubelist, const=const, model=model))
    try:
        cubelist.extract_cube(model.dens)
    except ConMisErr:
        cubelist.append(air_density(cubelist, const=const, model=model))
    try:
        cubelist.extract_cube(model.ghgt)
    except ConMisErr:
        cubelist.append(
            geopotential_height(cubelist, const=const, model=model)
        )


@update_metadata(units="m2 s-2")
def geopotential_height(cubelist, const=None, model=um):
    """
    Get geopotential height from the given cube list.

    If not present, the altitude coordinate is transformed into a cube.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes containing temperature.
    const: aeolus.const.const.ConstContainer, optional
        Must have `gravity` as an attribute.
        If not given, an attempt is made to retrieve it from cube attributes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of geopotential height.
    """
    try:
        return cubelist.extract_cube(model.ghgt)
    except ConMisErr:
        try:
            cube_w_height = cubelist.extract(
                _dim_constr(model.z, strict=False)
            )[0]
            if const is None:
                const = cube_w_height.attributes["planet_conf"]
            g_hgt = coord_to_cube(cube_w_height, model.z) * const.gravity
            g_hgt.attributes = {
                k: v
                for k, v in cube_w_height.attributes.items()
                if k != "STASH"
            }
            ensure_bounds(g_hgt, [model.z])
            g_hgt.rename(model.ghgt)
            return g_hgt
        except ConMisErr:
            _msg = f"No cubes in \n{cubelist}\nwith {model.z} as a coordinate."
            raise MissingCubeError(_msg)


def flux(cubelist, quantity, axis, weight_by_density=True, model=um):
    r"""
    Calculate horizontal or vertical flux of some quantity.

    .. math::
        F_{x} = u (\rho q),
        F_{y} = v (\rho q),
        F_{z} = w (\rho q)

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes.
    quantity: str or iris.Constraint
        Quantity (present in the cube list).
    axis: str
        Axis of the flux component (x|y|z)
    weight_by_density: bool, optional
        Multiply by air density (must be present in the input cube list).
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of a flux component with the same dimensions as input cubes.
    """
    if axis.lower() == "x":
        u = cubelist.extract_cube(model.u)
    elif axis.lower() == "y":
        u = cubelist.extract_cube(model.v)
    elif axis.lower() == "z":
        u = cubelist.extract_cube(model.w)
    q = cubelist.extract_cube(quantity)
    fl = u * q
    if weight_by_density:
        fl *= cubelist.extract_cube(model.dens)
    fl.rename(f"flux_of_{quantity}_along_{axis.lower()}_axis")
    return fl


@update_metadata(units="W m-2")
def toa_cloud_radiative_effect(cubelist, kind, model=um):
    r"""
    Calculate domain-average TOA cloud radiative effect (CRE).

    .. math::
        CRE_{TOA} = F_{up,clear-sky} - F_{up,all-sky}

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes
    kind: str
        Shortwave ('sw'), longwave ('lw'), or 'total' CRE

    Returns
    -------
    iris.cube.Cube
        Cube of CRE.
    """
    name = f"toa_cloud_radiative_effect_{kind}"
    if kind == "sw":
        all_sky = model.toa_osr
        clr_sky = model.toa_osr_cs
    elif kind == "lw":
        all_sky = model.toa_olr
        clr_sky = model.toa_olr_cs
    elif kind == "total":
        sw = toa_cloud_radiative_effect(cubelist, "sw", model=model)
        lw = toa_cloud_radiative_effect(cubelist, "lw", model=model)
        cre = sw + lw
        cre.rename(name)
        return cre

    cube_clr = cubelist.extract_cube(clr_sky)
    cube_all = cubelist.extract_cube(all_sky)
    cre = cube_clr - cube_all
    cre.rename(name)
    return cre


@update_metadata(name="toa_net_downward_energy_flux", units="W m-2")
def toa_net_energy(cubelist, model=um):
    """
    Calculate domain-average TOA energy flux.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of total TOA downward energy flux.
    """
    varnames = [
        model.toa_isr,
        model.toa_osr,
        model.toa_olr,
    ]
    terms = cubelist.extract(varnames)
    if len(terms) != 3:
        raise MissingCubeError(
            f"{varnames} required for TOA energy balance"
            " are missing from cubelist:\n{cubelist}"
        )
    terms_ave = []
    for cube in terms:
        terms_ave.append(cube)
    toa_net = terms_ave[0] - terms_ave[1] - terms_ave[2]
    return toa_net


@update_metadata(name="surface_net_downward_energy_flux", units="W m-2")
def sfc_net_energy(cubelist, model=um):
    """
    Calculate domain-average surface energy flux.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input cubes with net LW and SW, sensible and latent surface fluxes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of total surface downward energy flux.
    """
    net_down_lw = cubelist.extract_cube(model.sfc_net_down_lw)
    net_down_sw = cubelist.extract_cube(model.sfc_net_down_sw)
    shf = cubelist.extract_cube(model.sfc_shf)
    lhf = cubelist.extract_cube(model.sfc_lhf)
    sfc_net = net_down_lw + net_down_sw - shf - lhf
    return sfc_net


@const_from_attrs()
@update_metadata(name="surface_net_downward_water_flux", units="mm h-1")
def sfc_water_balance(cubelist, const=None, model=um):
    """
    Calculate domain-average precipitation minus evaporation.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes.
    const: aeolus.const.const.ConstContainer, optional
        Must have a scalar cube of `condensible_density` as an attribute.
        If not given, attempt is made to retrieve it from cube attributes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of total surface downward water flux (P-E).
    """
    try:
        evap = cubelist.extract_cube(model.sfc_evap)
    except ConMisErr:
        try:
            lhf = cubelist.extract_cube(model.sfc_lhf)
            evap = lhf / const.condensible_heat_vaporization
            evap /= const.condensible_density
        except (KeyError, ConMisErr):
            raise MissingCubeError(
                f"Cannot retrieve evaporation from\n{cubelist}"
            )
    try:
        precip = cubelist.extract_cube(model.ppn)
        precip /= const.condensible_density
    except ConMisErr:
        precip = precip_sum(cubelist, ptype="total", const=const, model=model)
    precip.convert_units("mm h-1")
    evap.convert_units("mm h-1")
    net = precip - evap
    return net


@const_from_attrs()
def precip_sum(cubelist, ptype="total", const=None, model=um):
    """
    Calculate a sum of different types of precipitation [:math:`mm~day^{-1}`].

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes.
    ptype: str, optional
        Precipitation type (total|stra|conv|rain|snow).
    const: aeolus.const.const.ConstContainer, optional
        Must have a scalar cube of `condensible_density` as an attribute.
        If not given, attempt to retrieve it from cube attributes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Sum of cubes of precipitation with units converted to mm per day.
    """
    try:
        varnames = _precip_name_mapping(model=model)[ptype]
    except KeyError:
        raise ArgumentError(f"Unknown ptype={ptype}")
    if len(cubelist.extract(varnames)) == 0:
        raise MissingCubeError(
            f"{varnames} required for ptype={ptype} "
            "are missing from cubelist:\n{cubelist}"
        )
    precip = 0.0
    for varname in varnames:
        try:
            cube = cubelist.extract_cube(varname)
            if varname in [model.cv_rain, model.cv_snow]:
                cube = cube.copy(data=cube.data.filled(fill_value=0))
            precip += cube
        except ConMisErr:
            pass
    precip /= const.condensible_density
    precip.convert_units("mm day-1")
    precip.rename(f"{ptype}_precip_rate")
    return precip


@update_metadata(name="heat_redistribution_efficiency", units="1")
def heat_redist_eff(cubelist, region_a, region_b, model=um):
    r"""
    Heat redistribution efficiency (Leconte et al. 2013).

    .. math::
        \eta=\frac{OLR_{TOA,night}}{OLR_{TOA,day}}

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes.
    region_a: aeolus.region.Region
        First region (usually nightside).
    region_b: aeolus.region.Region
        Second region (usually dayside).
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of eta parameter with collapsed spatial dimensions.
    """
    toa_olr = cubelist.extract_cube(model.toa_olr)
    toa_olr_a = spatial_mean(toa_olr.extract(region_a.constraint))
    toa_olr_b = spatial_mean(toa_olr.extract(region_b.constraint))
    eta = toa_olr_a / toa_olr_b
    return eta


@update_metadata(name="toa_effective_temperature", units="K")
def toa_eff_temp(cubelist, model=um):
    r"""
    Calculate effective temperature from TOA OLR.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of :math:`T_{eff}`.
    """
    toa_olr = cubelist.extract_cube(model.toa_olr)
    sbc = init_const().stefan_boltzmann
    t_eff = (toa_olr / sbc) ** 0.25
    return t_eff


@update_metadata(name="normalised_greenhouse_effect_parameter", units="1")
def ghe_norm(cubelist, model=um):
    r"""
    Normalised greenhouse effect parameter.

    .. math::
        GHE = 1 - \left(\frac{T_{eff}}{T_{sfc}}\right)^4

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of greenhouse effect parameter.
    """
    t_sfc = cubelist.extract_cube(model.t_sfc)
    t_eff = toa_eff_temp(cubelist, model=model)
    out = (t_eff / t_sfc) ** 4
    out = out.copy(data=1 - out.data)
    return out


@const_from_attrs()
@update_metadata(name="greenhouse_effect_parameter", units="K")
def greenhouse_effect(cubelist, kind="all_sky", const=None, model=um):
    r"""
    Calculate the greenhouse effect [K].

    .. math::
        GHE = T_{sfc} - \left(\frac{T_{eff}}{T_{sfc}}\right)^4

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes.
    kind: str, optional
        Type of GHE:  "all_sky" or "clear_sky"
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of greenhouse effect parameter.
    """
    t_sfc = cubelist.extract_cube(model.t_sfc)
    if kind == "all_sky":
        toa_olr = cubelist.extract_cube(model.toa_olr)
    elif kind == "clear_sky":
        toa_olr = cubelist.extract_cube(model.toa_olr_cs)
    ghe = t_sfc - (toa_olr / const.stefan_boltzmann) ** 0.25
    return ghe


@const_from_attrs()
@update_metadata(name="bond_albedo", units="1")
def bond_albedo(cubelist, const=None, model=um):
    r"""
    Bold albedo.

    .. math::
        4 \frac{OSR_{TOA}}{S_{0}}

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes.
    const: aeolus.const.const.ConstContainer, optional
        Must have a scalar cube of `solar_constant` as an attribute.
        If not given, attempt to retrieve it from cube attributes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of bond albedo.
    """
    toa_osr = cubelist.extract_cube(model.toa_osr)
    sc = const.solar_constant
    b_alb = 4 * toa_osr / sc
    return b_alb


@update_metadata(units="kg m-2")
def water_path(cubelist, kind="water_vapour", model=um):
    r"""
    Water vapour or condensate path, i.e. a vertical integral of a water phase.

    .. math::
        WP = \int_{z_{sfc}}^{z_{top}} \rho q dz

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes containing mixing ratio and air density.
    kind: str, optional
        Short name of the water phase to be integrated.
        Options:
        water_vapour (default) | liquid_water | ice_water | cloud_water
        where `cloud_water` is the sum of liquid and ice phases.
    model: aeolus.model.Model, optional
        Model class with relevant coordinate names.
        `model.z` is used as a vertical coordinate for integration.

    Returns
    -------
    iris.cube.Cube
        Cube of water path with collapsed vertical dimension.
    """
    if kind == "water_vapour":
        q = cubelist.extract_cube(model.sh)
    elif kind == "liquid_water":
        q = cubelist.extract_cube(model.cld_liq_mf)
    elif kind == "ice_water":
        q = cubelist.extract_cube(model.cld_ice_mf)
    elif kind == "cloud_water":
        q = cubelist.extract_cube(model.cld_liq_mf).copy()
        q += cubelist.extract_cube(model.cld_ice_mf)
    rho = cubelist.extract_cube(model.dens)
    wp = integrate(q * rho, model.z)
    wp.rename(f"{kind}_path")
    return wp


def dry_lapse_rate(cubelist, model=um):
    r"""
    Dry lapse rate, or the change of air temperature with altitude.

    .. math::
        \gamma = \partial T_{air} / \partial z

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes containing temperature.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of dT/dz.
    """
    temp = cubelist.extract_cube(model.temp)
    return d_dz(temp, model=model)


def horiz_wind_cmpnts(cubelist, model=um):
    """
    Extract u and v winds and interpolate v on u's grid if necessary.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        List of cubes with horizontal wind components
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    u, v: iris.cube.Cube
        Cubes of wind components.
    """
    # use relaxed extracting
    u = cubelist.extract(model.u)[0]
    v = cubelist.extract(model.v)[0]
    # interpolate v on u's grid if coordinates are different
    v = regrid_3d(v, u, model=model)
    return u, v


@const_from_attrs()
@update_metadata(name="local_superrotation_index", units="1")
def superrotation_index(cubelist, const=None, model=um):
    r"""
    Local superrotation index.

    .. math::
        s = \frac{m}{\Omega a^2} - 1,

        m = a cos\phi(\Omega a cos\phi + u)

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        List of cubes containing a cube of zonal velocity (u).
    const: aeolus.const.const.ConstContainer, optional
        Constainer with the relevant planetary constants.
        If not given, attempt is made to retrieve it from cube attributes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    s_idx: iris.cube.Cube
        Cubes of superrotation index.

    References
    ----------
    Read (1986), https://doi.org/10.1002/qj.49711247114
    """
    # Zonal velocity
    u = cubelist.extract_cube(model.u)
    # Radius of the planet
    r = const.radius
    r.convert_units("m")
    # Rotation rate
    omega = const.planet_rotation_rate

    lat_coord = u.coord(model.y).copy()
    lat_dim = u.coord_dims(model.y)[0]
    lat_coord.convert_units("radians")
    lat_cos_coord = _coord_cos(lat_coord)

    # Calculate intermediate terms
    omega_r_coslat = omega.data * r.data * lat_cos_coord
    omega_r_coslat.units = Unit("m s-1")
    r_coslat = r.data * lat_cos_coord
    r_coslat.units = Unit("m")
    inner_sum = add(u, omega_r_coslat, dim=lat_dim)

    # Calculate axial component of specific absolute angular momentum
    ang_mom = multiply(inner_sum, r_coslat, dim=lat_dim)

    # Final index
    s_idx = ang_mom / omega / (r**2)
    s_idx.convert_units("1")
    s_idx = s_idx.copy(data=s_idx.data - 1)
    return s_idx


@update_metadata(name="wind_speed", units="m s-1")
def wind_speed(*components):
    r"""
    Calculate the wind speed (magnitude of the wind vector).

    .. math::
        \sqrt{u^2 + v^2 + w^2}

    Parameters
    ----------
    args: iris.cube.Cube
        Cubes of u, v, w wind components.

    Returns
    -------
    iris.cube.Cube
    """
    out = sum(cube**2 for cube in components) ** 0.5
    return out


@update_metadata(name="atmosphere_hybrid_sigma_pressure_coordinate", units="1")
def sigma_p(cubelist, model=um):
    r"""
    Calculate sigma (normalised pressure coordinate) from a cube of pressure.

    .. math::
        \sigma = p / p_{sfc}

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.
    """
    pres_atm = cubelist.extract_cube(model.pres)
    pres_sfc = cubelist.extract_cube(model.p_sfc)
    out = pres_atm / pres_sfc
    return out


@const_from_attrs()
@update_metadata(name="zonal_mass_streamfunction", units="kg s^-1")
def zonal_mass_streamfunction(cubelist, const=None, model=um):
    r"""
    Calculate mean zonal mass streamfunction.

    .. math::
        \Psi_Z=2\pi a\int_{z_{sfc}}^{z_{top}}\overline{\rho}^*\overline{u}^* dz

    References
    ----------
    Haqq-Misra & Kopparapu (2015), eq. 5;
    Hartmann (1994), Global Physical Climatology, eq. 6.21

    Examples
    --------
    >>> lat_band_constr = iris.Constraint(
        latitude=lambda x: -30 <= x.point <= 30, longitude=lambda x: True
    )
    >>> mzsf = zonal_mass_streamfunction(cubes.extract(lat_band_constr))
    """
    streamf_const = 2 * np.pi * const.radius

    u = cubelist.extract_cube(model.u)
    u_tm = time_mean(u, model=model)
    u = -1 * (u_tm - zonal_mean(u_tm, model=model))

    if u.coord(axis="z").units.is_convertible("m"):
        rho = cubelist.extract_cube(model.dens).copy()
        rho = time_mean(rho, model=model)
        rho.coord(model.z).bounds = None
        u.coord(model.z).bounds = None
        integrand = u * rho
        # integrand = reverse(integrand, model.z)
        # print(integrand.coord(um.z))
        res = cumsum(integrand, "z", axis_weights=True, model=model)
        # res = cumsum(integrand, "z", axis_weights=True, model=model)
        # res = reverse(res, model.z)

    elif u.coord(axis="z").units.is_convertible("Pa"):
        integrand = u
        res = cumsum(integrand, "z", axis_weights=True, model=model)
        res /= const.gravity

    res *= streamf_const
    return res


@const_from_attrs()
@update_metadata(name="meridional_mass_streamfunction", units="kg s^-1")
def meridional_mass_streamfunction(cubelist, const=None, model=um):
    r"""
    Calculate the mean meridional mass streamfunction.

    * In height coordinates

    .. math::
        \Psi_M = - 2\pi cos\phi a \int_{z_{sfc}}^{z_{top}}\overline{\rho v} dz

    * In pressure coordinates

    .. math::
        \Psi_M = 2\pi cos\phi a \int_{0}^{p_{sfc}}\overline{v} dp / g

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input cubelist.
    const: aeolus.const.const.ConstContainer, optional
        If not given, constants are attempted to be retrieved from
        attributes of a cube in the cube list.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube with collapsed spatial dimensions.

    References
    ----------
    Haqq-Misra & Kopparapu (2015), eq. 4;
    Vallis (2017)

    Examples
    --------
    >>> from aeolus.calc import meridional_mass_streamfunction, time_mean
    >>> from aeolus.const import init_const
    >>> from aeolus.model import um
    >>> earth_constants = init_const("earth")
    >>> cubes = iris.cube.CubeList([time_mean(cube) for cube in input_cubes])
    >>> mmsf = meridional_mass_streamfunction(cubes, const=earth_constants)
    """
    v = cubelist.extract_cube(model.v)
    v = zonal_mean(v, model=model)
    if v.coord(model.z).units.is_convertible("m"):
        rho = zonal_mean(cubelist.extract_cube(model.dens), model=model)
        rho.coord(model.z).bounds = None
        v.coord(model.z).bounds = None
        # Reverse the coordinate to start from the top (where p=0 or z=z_top)
        # TODO: check if the coordinate is ascending or descending
        integrand = reverse(v * rho, model.z)
        res = -1 * cumsum(integrand, "z", axis_weights=True, model=model)
        # Reverse the result back
        res = reverse(res, model.z)
    elif v.coord(model.z).units.is_convertible("Pa"):
        res = cumsum(v, "z", axis_weights=True, model=model)
        res /= const.gravity
    # Calculate the constant: 2 pi cos\phi a
    streamf_const = 2 * np.pi * const.radius * lat_cos(res, model=model)
    res *= streamf_const
    return res


@const_from_attrs()
def wind_rot_div(u, v, truncation=None, const=None, model=um):
    """
    Split the horizontal wind field into divergent and rotational parts.

    The Helmholtz decomposition method uses the `windspharm` library:
    https://ajdawson.github.io/windspharm/latest/

    Parameters
    ----------
    u: iris.cube.Cube
        Eastward wind.
    v: iris.cube.Cube
        Northward wind.
    truncation: int
        Truncation for the spherical harmonic computation
        (See windspharm docs for details).
    const: aeolus.const.const.ConstContainer, optional
        If not given, constants are attempted to be retrieved from
        attributes of a cube in the cube list.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    out: dict
        Dictionary of cubes of:
          - input wind components (for convenience),
          - divergent components,
          - rotational components,
          - zonal mean rotational components,
          - zonal eddy rotational components.

    References
    ----------
    Hammond and Lewis (2021), https://doi.org/10.1073/pnas.2022705118.
    """
    from windspharm.iris import VectorWind

    vec = VectorWind(u, v, rsphere=const.radius.data)
    div_cmpnt_u, div_cmpnt_v, rot_cmpnt_u, rot_cmpnt_v = vec.helmholtz(
        truncation=truncation
    )
    out = {}
    out["u_total"] = u
    out["v_total"] = v
    out["u_div"] = reverse(div_cmpnt_u, model.y)
    out["v_div"] = reverse(div_cmpnt_v, model.y)
    out["u_rot"] = reverse(rot_cmpnt_u, model.y)
    out["v_rot"] = reverse(rot_cmpnt_v, model.y)

    for cmpnt in ["u", "v"]:
        rot_cmpnt = out[f"{cmpnt}_rot"]
        out[f"{cmpnt}_rot_zm"] = preserve_shape(zonal_mean)(rot_cmpnt)
        out[f"{cmpnt}_rot_zm"].rename(f"zonal_mean_of_{rot_cmpnt.name()}")
        out[f"{cmpnt}_rot_eddy"] = rot_cmpnt - out[f"{cmpnt}_rot_zm"]
        out[f"{cmpnt}_rot_eddy"].rename(
            f"zonal_deviation_of_{rot_cmpnt.name()}"
        )
    return out



"""Bells and whistles."""

from iris.experimental.representation import CubeListRepresentation


class ReprAtmoSimBase:
    """
    Produce representations of AtmoSimBase-like instance.

    This includes:
    * ``str_repr``: provides __repr__ or __str__ view as a string
    * ``html_repr``: as an HTML object, available in Jupyter notebooks.
        Specifically, this is presented as an HTML table.
    """

    _template = """
<style>
  table.aeolus {{
      white-space: pre;
      border: 1px solid;
      border-color: #f9f9ef;
      font-family: monaco, monospace;
  }}
  th.aeolus {{
      background: #084469;
      color: #fefefe;
      border-left: 1px solid;
      border-color: #0a507a;
      font-size: 1.05em;
      min-width: 50px;
      max-width: 125px;
  }}
  .aeolus-la {{
      text-align: left !important;
      white-space: pre;
  }}
  .aeolus-ra {{
      text-align: right !important;
      white-space: pre;
  }}
</style>
<table class="aeolus" id="{id}">
    {header}
    {content}
</table>
        """

    def __init__(self, atmosim):
        """
        Initialise ReprAtmoSimBase.

        Parameters
        ----------
        atmosim: aeolus.core.AtmoSimBase
            AtmoSimBase-like instance
        """
        self._cls_m = atmosim.__module__
        self._cls_n = atmosim.__class__.__name__
        self.as_id = id(atmosim)
        self.atmosim = atmosim
        self._cubes = self.atmosim._cubes
        self.name = f"{self._cls_m} {self._cls_n} '{self.atmosim.name}'"
        self._copy_attrs = [
            "description",
            "planet",
            "model",
            "model_type",
            "timestep",
            "domain",
            "vert_coord",
        ]
        self._max_len = max([len(i) for i in self._copy_attrs])
        for attr in self._copy_attrs:
            setattr(self, attr, getattr(atmosim, attr))

    def str_repr(self, short=False):
        """Represent AtmoSimBase as string."""
        summary = []
        summary.append(f"{self.name} [{len(self._cubes)} cubes]")
        # Short option - for __repr__
        if short:
            return " ".join(summary)
        # Long option - for __str__
        for attr in self._copy_attrs:
            if getattr(self, attr) is not None:
                summary.append(
                    f"{attr:>{self._max_len}} | {getattr(self, attr)}"
                )
        if len(self._cubes) > 0:
            summary.append(f"{self._cubes!s}\n")
        return "\n".join(summary)

    def _make_header(self):
        top_cell = f'<th class="aeolus aeolus-la" colspan="2">{self.name}</th>'
        cells = ['<tr class="aeolus">', top_cell, "</tr>"]
        return "\n".join(cells)

    def _make_content(self):
        cells = []
        for attr in self._copy_attrs:
            if getattr(self, attr) is not None:
                value = getattr(self, attr)
                cells.append('<tr class="aeolus">')
                cells.append(f'<td class="aeolus aeolus-ra">{attr}</td>')
                cells.append(f'<td class="aeolus aeolus-la">{value!s}</td>')
                cells.append("</tr>")
        if len(self._cubes) > 0:
            # List cubes
            cl_repr_html = CubeListRepresentation(self._cubes).repr_html()
            cells.append('<tr class="aeolus">')
            cells.append('<td class="octant aeolus-ra">Cubes</td>')
            cells.append(f'<td class="aeolus aeolus-la">{cl_repr_html}</td>')
            cells.append("</tr>")

        return "\n".join(cells)

    def html_repr(self):
        """HTML representation used in Jupyter Notebooks."""
        header = self._make_header()
        content = self._make_content()
        return self._template.format(
            id=self.as_id, header=header, content=content
        )



"""Post-process model output by interpolating it to a common grid."""

# Standard library
from functools import partial
import os
from pathlib import Path
from time import time

# External libraries
from .const import add_planet_conf_to_cubes, init_const
from .io import create_dummy_cube, save_cubelist
from .lfric import (
    add_equally_spaced_height_coord,
    add_um_height_coord,
    fix_time_coord,
    load_lfric_raw,
    simple_regrid_lfric,
)
from .log import create_logger
from .subset import unique_cubes

os.environ["PROJ_IGNORE_CELESTIAL_BODY"] = "YES"


def chunker(seq, size):
    """Iterate over a sequence in chunks."""
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


def process_lfric(
    inpdir: str,
    outdir: str,
    label: str,
    planet: str,
    c_num: str,
    ref_cube: str,
    level_height: str,
    top_height: float,
    time_prof: str,
    file_chunk_size: int,
    nlat: int,
    nlon: int,
) -> Path:
    """Post-process LFRic output."""
    t0 = time()
    L = create_logger(Path(__file__))
    # Input directory
    if inpdir:
        inpdir = Path(inpdir)
    else:
        try:
            inpdir = Path(os.environ["AEOLUS_PROJ_INP_DIR"]) / label / c_num
            L.info(f"{inpdir=}")
        except KeyError:
            err_msg = (
                "Environment variable AEOLUS_PROJ_INP_DIR not defined."
                "Define it or use the --inpdir argument explicitly."
            )
            L.critical(err_msg)
            return
    # Create a subdirectory for processed data
    if outdir:
        outdir = Path(outdir)
    else:
        try:
            outdir = Path(os.environ["AEOLUS_PROJ_OUT_DIR"]) / label / c_num
            L.info(f"{outdir=}")
        except KeyError:
            err_msg = (
                "Environment variable AEOLUS_PROJ_OUT_DIR not defined."
                "Define it or use the --outdir argument explicitly."
            )
            L.critical(err_msg)
            return
    outdir.mkdir(parents=True, exist_ok=True)
    # Global attributes of the output file
    gl_attrs = {
        "name": label,
        "planet": planet,
        "processed": "True",
    }
    # Create a dummy cube with a target grid
    tgt_cube = create_dummy_cube(nlat=nlat, nlon=nlon, pm180=True)

    # Height coordinate
    if level_height == "uniform":
        add_levs = partial(
            add_equally_spaced_height_coord, model_top_height=top_height
        )
    elif level_height.startswith("um_"):
        add_levs = partial(
            add_um_height_coord,
            path_to_levels_file=Path(os.environ["AEOLUS_PROJ_VERT_DIR"])
            / level_height.replace("um_", "vertlevs_"),
        )
    else:
        raise ValueError(f"{level_height=} is not valid.")

    def _combi_callback(cube, field, filename):
        [
            fix_time_coord(cube, field, filename),
            add_levs(cube, field, filename),
        ]

    # File names
    if time_prof == "inst":
        file_stem = "lfric*diag"
        drop_coord = ["forecast_reference_time"]
    elif time_prof == "mean":
        file_stem = "lfric_averages"
        drop_coord = []
    else:
        raise ValueError(f"{time_prof=} is not valid.")

    # Compile the list of files
    fnames = sorted(
        inpdir.glob(f"*/*{c_num}*/{file_stem}.nc"),
        key=lambda x: int(x.parent.parent.name),
    )
    if len(fnames) == 0:
        L.critical("No files found!")
        return
    elif len(fnames) == 1:
        L.info(f"fnames({len(fnames)}) = {fnames[0]}")
    else:
        L.info(f"fnames({len(fnames)}) = {fnames[0]} ... {fnames[-1]}")

    # Load the data in chunks
    for _ichunk, file_batch in enumerate(chunker(fnames, file_chunk_size)):
        L.info(
            f"Loading {file_chunk_size} files ="
            f" {file_batch[0]} ... {file_batch[-1]}"
        )
        # Make a label depending on the chunk
        chunk_label = f"_{int(file_batch[0].parent.parent.name):03d}"
        chunk_label += f"-{int(file_batch[-1].parent.parent.name):03d}"
        gl_attrs["source_files"] = ", ".join(
            [str(Path(*f.parts[-3:])) for f in file_batch]
        )
        # Load the data
        cl_raw = load_lfric_raw(
            file_batch, callback=_combi_callback, drop_coord=drop_coord
        )
        if len(cl_raw) == 0:
            L.critical("Files are empty!")
            continue
        cubes_to_regrid = unique_cubes(cl_raw)
        if len(w_cubes := cubes_to_regrid.extract("w_in_wth")) == 2:
            cubes_to_regrid.remove(w_cubes[-1])
        for cube in cubes_to_regrid:
            if cube.units == "ms-1":
                cube.units = "m s-1"
        # Regrid all cubes
        cl_proc = simple_regrid_lfric(
            cubes_to_regrid, tgt_cube=tgt_cube, ref_cube_constr=ref_cube
        )
        const = init_const(
            planet, directory=Path(os.environ["AEOLUS_PROJ_CONST_DIR"])
        )
        add_planet_conf_to_cubes(cl_proc, const=const)

        # Write the data to a netCDF file
        fname_out = (
            outdir
            / f"{label}_{c_num}_{time_prof}_{chunk_label}_regr.nc".lower()
        )
        save_cubelist(cl_proc, fname_out, **gl_attrs)
        L.success(f"Saved to {fname_out}")
    L.info(f"Execution time: {time() - t0:.1f}s")
    return fname_out



"""Metadata-related functionality."""

from collections.abc import Iterable
from dataclasses import is_dataclass
import functools

import cf_units
from iris.analysis import _dimensional_metadata_comparison
from iris.cube import Cube, CubeList
from iris.util import broadcast_to_shape

from .exceptions import ArgumentError  # , _warn


def const_from_attrs(strict=True):
    """Get constants from the cube attributes if not passed explicitly."""

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            const = kwargs.pop("const", None)
            if const is None:
                for arg in args:
                    if isinstance(arg, Cube):
                        const = arg.attributes.get("planet_conf")
                    elif isinstance(arg, CubeList):
                        try:
                            const = arg[0].attributes.get("planet_conf")
                        except IndexError:
                            const = None
            if is_dataclass(const):
                kwargs.update(const=const)
            else:
                msg = (
                    "`const` has to be the function argument "
                    "or in the cube attributes."
                )
                if strict:
                    raise ArgumentError(msg)
                # else:
                #     _warn(msg)
            # Call the decorated function
            out = func(*args, **kwargs)
            return out

        return wrapper

    return decorator


def copy_doc(original):
    """Copy docstring from another function."""

    def wrapper(func):
        func.__doc__ = original.__doc__
        return func

    return wrapper


def preserve_shape(func):
    """Preserve shape of the output array."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        cube_in = args[0]
        if isinstance(cube_in, Cube):
            orig_shape = cube_in.shape
        else:
            raise ArgumentError(
                "`preserve_shape` decorator requires the first argument"
                "of the function and its output to be a cube."
            )
        # Call the decorated function
        cube_out = func(*args, **kwargs)
        out_name = cube_out.name()
        cell_methods = cube_out.cell_methods
        dim_map = []
        for ndim, _ in enumerate(orig_shape):
            for pair in _dimensional_metadata_comparison(cube_out, cube_in)[
                "not_equal"
            ]:
                if ndim not in cube_in.coord_dims(pair[0]):
                    dim_map.append(ndim)
        bc_data = broadcast_to_shape(
            cube_out.data, orig_shape, sorted(set(dim_map))
        )
        cube_out = cube_in.copy(data=bc_data)
        cube_out.rename(out_name)
        cube_out.cell_methods = cell_methods
        return cube_out

    return wrapper


def update_metadata(name=None, units=None, attrs=None):
    """
    Update metadata of a cube returned by a function.

    If units are "unknown", the units are reset to the given units,
    otherwise `convert_units()` is used.
    """

    def _update(cube):
        """Update name, convert units and update attributes."""
        if isinstance(name, str):
            cube.rename(name)
        if isinstance(units, (str, cf_units.Unit)):
            if cube.units == "unknown":
                cube.units = units
            else:
                cube.convert_units(units)
        if isinstance(attrs, dict):
            cube.attributes.update(attrs)

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Call the decorated function
            out = func(*args, **kwargs)
            if isinstance(out, Cube):
                _update(out)
            elif isinstance(out, Iterable):
                [_update(cube) for cube in out]
            return out

        return wrapper

    return decorator



"""Subsetting variables over geographical regions."""

from iris import Constraint
from iris.analysis.cartography import wrap_lons

from .exceptions import BoundaryError
from .model import um
from .plot.text import fmt_lonlat

__all__ = ("Region",)


class BoundsRect:
    """Bounding longitudes and latitudes of a given lon-lat rectangle."""

    _side_names = ("west", "east", "south", "north")

    def __init__(
        self, west_bound, east_bound, south_bound, north_bound, model=um
    ):
        """Initialise BoundsRect."""
        self.west = west_bound
        self.east = east_bound
        self.south = south_bound
        self.north = north_bound
        self.model = model
        self.west_coord = model.x
        self.east_coord = model.x
        self.south_coord = model.y
        self.north_coord = model.y
        if self.south > self.north:
            raise BoundaryError(
                f"South boundary value ({self.south}) should be less than"
                f" north ({self.north})"
            )
        self.sides = [
            (name, getattr(self, f"{name}_coord")) for name in self._side_names
        ]

    def __repr__(self):  # noqa
        return (
            f"BoundsRect(west={self.west}, "
            f"east={self.east}, "
            f"south={self.south}, "
            f"north={self.north})"
        )


class Region:
    """
    Rectangular geographical region.

    Attributes
    ----------
    name : str
        The region's name
    description : str
        A description of the region
    constraint: iris.Constraint
        A constraint object associated with the region
    """

    def __init__(
        self,
        west_bound,
        east_bound,
        south_bound,
        north_bound,
        name="",
        description="",
        model=um,
    ):
        """
        Instantiate a `Region` object.

        Parameters
        ----------
        west_bound, east_bound, south_bound, north_bound : scalar, optional
            The western, eastern, southern, and northern boundaries,
            respectively, of the region.
        name: str, optional
            The region's name.
        description : str, optional
            A description of the region.
        model: aeolus.model.Model, optional
            Model class with a relevant coordinate names.
        """
        self.name = name
        self.description = description
        self.model = model

        self.bounds = BoundsRect(
            west_bound, east_bound, south_bound, north_bound, model
        )
        self._sides = self.bounds.sides

        self.lon_size = abs(self.bounds.east - self.bounds.west)
        self.lat_size = self.bounds.north - self.bounds.south

    def __repr__(self):  # noqa
        txt = (
            f"Geographical region '{self.name}' (west={self.bounds.west}, "
            f"east={self.bounds.east}, south={self.bounds.south},"
            f" north={self.bounds.north})"
        )
        if self.description:
            txt += "\n\n"
            txt += self.description
        return txt

    def __getitem__(self, index):  # noqa
        return {
            "value": getattr(self.bounds, self._sides[index][0]),
            "name": self._sides[index][0],
            "coord": self._sides[index][1],
        }

    def _perpendicular_side_limits(self, side):
        """Get the min and max values of the region boundary perpendicular.

        to the given one.
        """
        if side in ["west", "east"]:
            coord_name = self.model.y
            _min, _max = self.bounds.south, self.bounds.north
        elif side in ["south", "north"]:
            coord_name = self.model.x
            _min, _max = self.bounds.west, self.bounds.east
        else:
            raise BoundaryError(f"Boundary name '{side}' is not valid")
        return coord_name, (_min, _max)

    def to_str(self, sep="_"):  # noqa
        return sep.join([fmt_lonlat(i["value"], i["coord"]) for i in self])

    @classmethod
    def from_cube(
        cls,
        cube,
        name=None,
        margin=None,
        margin_units="points",
        shift_lons=False,
        model=um,
    ):
        """
        Create a Region from limits of longitude and latitude of the cube.

        Parameters
        ----------
        cube: iris.cube.Cube
            Source cube.
        name: str, optional
            Name for the region.
            If not given, created automatically from `cube`'s name.
        margin: scalar, optional
            Number of points or degrees for a region smaller than the cube.
        margin_units: str, optional
            Units of margin. Can be "points" or "degrees".
        shift_lons: bool, optional
            Shift longitudes to -180...180.
        model: aeolus.model.Model, optional
            Model class with a relevant coordinate names.

        Returns
        -------
        aeolus.region.Region
        """
        if name is None:
            name = f"extent_of_{cube.name()}"
        lons = cube.coord(model.x).points
        if shift_lons:
            lons = sorted(wrap_lons(lons, -180, 360))
        lats = cube.coord(model.y).points
        idx0, idx1 = 0, -1
        if margin is not None:
            if margin_units == "points":
                idx0 += margin
                idx1 -= margin
                lon0 = lons[idx0]
                lon1 = lons[idx1]
                lat0 = lats[idx0]
                lat1 = lats[idx1]
            else:
                lon0 = lons[idx0] + margin
                lon1 = lons[idx1] - margin
                lat0 = lats[idx0] + margin
                lat1 = lats[idx1] - margin
        else:
            lon0 = lons[idx0]
            lon1 = lons[idx1]
            lat0 = lats[idx0]
            lat1 = lats[idx1]

        return cls(lon0, lon1, lat0, lat1, name=name)

    @property
    def constraint(self):  # noqa
        cnstr = Constraint(
            latitude=lambda x: self.bounds.south
            <= x.point
            <= self.bounds.north
        )
        if self.bounds.west < self.bounds.east:
            # Western boundary is to the west
            cnstr &= Constraint(
                longitude=lambda x: self.bounds.west
                <= x.point
                <= self.bounds.east
            )
        else:
            # Region wrapping around dateline (180deg)
            cnstr &= Constraint(
                longitude=lambda x: (self.bounds.west <= x.point)
                or (x.point <= self.bounds.east)
            )
        return cnstr

    def add_to_ax(self, ax, **kwargs):
        """Add a Rectangle patch to matplotlib axes."""
        from matplotlib.patches import Rectangle  # noqa

        xy = (self.bounds.west, self.bounds.south)
        width = self.lon_size
        height = self.lat_size
        ax.add_patch(Rectangle(xy, width, height, **kwargs))



"""Functionality related to coordinates of cubes."""

from datetime import timedelta

from cartopy.util import add_cyclic_point
import dask.array as da
from iris import Constraint
import iris.analysis
import iris.analysis.cartography
from iris.analysis.cartography import wrap_lons
from iris.coord_categorisation import _months_in_season, add_categorised_coord
from iris.coords import AuxCoord, DimCoord
from iris.cube import Cube, CubeList
from iris.exceptions import CoordinateNotFoundError as CoNotFound
import iris.fileformats
from iris.util import (
    broadcast_to_shape,
    guess_coord_axis,
    is_regular,
    promote_aux_coord_to_dim_coord,
    squeeze,
)
import numpy as np

from .const import get_planet_radius
from .exceptions import ArgumentError, BadCoordinateError, NotFoundError, _warn
from .meta import const_from_attrs
from .model import um

__all__ = (
    "CoordContainer",
    "add_binned_coord",
    "add_cyclic_point_to_cube",
    "add_planet_calendar",
    "area_weights_cube",
    "check_coords",
    "coarsen_cube",
    "coord_delta_to_cube",
    "coord_to_cube",
    "ensure_bounds",
    "get_cube_datetimes",
    "get_cube_rel_days",
    "get_dim_coord",
    "get_xy_coords",
    "isel",
    "interp_cube_from_height_to_pressure_levels",
    "interp_cubelist_from_height_to_pressure_levels",
    "interp_to_cube_time",
    "nearest_coord_value",
    "not_equal_coord_axes",
    "regrid_3d",
    "replace_z_coord",
    "roll_cube_0_360",
    "roll_cube_pm180",
    "vertical_cross_section_area",
    "volume_weights_cube",
)


def _cell_bounds(points, bound_position=0.5):
    """
    Calculate coordinate cell boundaries.

    Taken from SciTools iris package.

    Parameters
    ----------
    points: numpy.ndarray
        One-dimensional array of uniformy spaced values of shape (M,)
    bound_position: bool, optional
        The desired position of the bounds relative to the position
        of the points.

    Returns
    -------
    bounds: numpy.ndarray
        Array of shape (M+1,)

    Examples
    --------
    >>> a = np.arange(-1, 2.5, 0.5)
    >>> a
    array([-1. , -0.5,  0. ,  0.5,  1. ,  1.5,  2. ])
    >>> cell_bounds(a)
    array([-1.25, -0.75, -0.25,  0.25,  0.75,  1.25,  1.75,  2.25])

    See Also
    --------
    aeolus.coord._cell_centres
    """
    assert points.ndim == 1, "Only 1D points are allowed"
    diffs = np.diff(points)
    if not np.allclose(diffs, diffs[0]):
        _warn("_cell_bounds() assumes uniformly spaced points")
    delta = diffs[0] * bound_position
    bounds = np.concatenate([[points[0] - delta], points + delta])
    return bounds


def _cell_centres(bounds, bound_position=0.5):
    """
    Calculate coordinate cell centres.

    Taken from SciTools iris package.

    Parameters
    ----------
    bounds: numpy.ndarray
        One-dimensional array of cell boundaries of shape (M,)
    bound_position: bool, optional
        The desired position of the bounds relative to the position
        of the points.

    Returns
    -------
    centres: numpy.ndarray
        Array of shape (M-1,)

    Examples
    --------
    >>> a = np.arange(-1, 3., 1.)
    >>> a
    array([-1,  0,  1,  2])
    >>> cell_centres(a)
    array([-0.5,  0.5,  1.5])

    See Also
    --------
    aeolus.coord._cell_bounds
    """
    assert bounds.ndim == 1, "Only 1D points are allowed"
    deltas = np.diff(bounds) * bound_position
    centres = bounds[:-1] + deltas
    return centres


def _is_longitude_global(lon_points):
    """Return True if array of longitudes covers the whole sphere."""
    dx = np.diff(lon_points)[0]  # assume regular grid
    case_0_360 = ((lon_points[0] - dx) <= 0) and ((lon_points[-1] + dx) >= 360)
    case_pm180 = ((lon_points[0] - dx) <= -180) and (
        (lon_points[-1] + dx) >= 180
    )
    return case_0_360 or case_pm180


class CoordContainer:
    """
    Coordinate container.

    Attributes
    ----------
    {x,y,z,t}: iris.coord.Coord
        Coordinates in the respective dimensions
    """

    def __init__(self, cubes, coord_check=False, model=um):
        """
        Instantiate an `AtmosFlow` object.

        Parameters
        ----------
        cubes: iris.cube.CubeList
            Atmospheric fields with necessary coordinates.
        coord_check: bool, optional
            Check if all cubes have the same set of coordinates.
        model: aeolus.model.Model, optional
            Model class with relevant coordinate and variable names.
        """
        if coord_check:
            check_coords(cubes)
        # self.x = cubes[0].coord(model.x)
        self.model = model
        for axis in ["x", "y", "z", "t"]:
            try:
                setattr(self, axis, cubes[0].coord(axis=axis, dim_coords=True))
            except CoNotFound:
                pass


def add_binned_coord(cube, coord_name, bins):
    """
    Bin coordinate points and add them as an auxiliary coordinate to a cube.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube with the given coordinate.
    coord_name: str or iris.coords.Coord
        Coordinate name.
    bins: array-like
        Coordinate bins.

    Returns
    -------
    iris.cube.Cube
    """
    cube_out = cube.copy()
    binned_points = np.digitize(cube_out.coord(coord_name).points, bins)
    binned_points = np.clip(binned_points, 0, len(bins) - 1)
    new_coord = AuxCoord(binned_points, long_name=f"{coord_name}_binned")
    cube_out.add_aux_coord(new_coord, cube_out.coord_dims(coord_name))
    return cube_out


def add_cyclic_point_to_cube(cube, coord=um.x):
    """
    Add a cyclic point to a cube and a corresponding coordinate.

    A wrapper for `cartopy.util.add_cyclic_point()` for iris cubes.

    Parameters
    ----------
    cube: iris.cube.Cube
        An n-dimensional cube of data to add a cyclic point to.
    coord: iris.coords.Coord or str
        A 1-dimensional coordinate which specifies the coordinate values for
        the dimension the cyclic point is to be added to. The coordinate
        values must be regularly spaced. Defaults to the "x"-coordinate.

    Returns
    -------
    iris.cube.Cube
        The cube with a cyclic point added.
    """
    the_coord = cube.coord(coord)
    dim = cube.coord_dims(the_coord)

    # TODO: use core_data()?
    cy_data, cy_coord_pnts = add_cyclic_point(
        cube.data, coord=the_coord.points, axis=dim[0]
    )

    dim_coords_and_dims = [
        (coord, cube.coord_dims(coord))
        for coord in cube.dim_coords
        if coord != the_coord
    ]
    dim_coords_and_dims.append((the_coord.copy(cy_coord_pnts), dim))

    aux_coords_and_dims = [
        (coord, cube.coord_dims(coord))
        for coord in cube.aux_coords
        if cube.coord_dims(coord) != dim
    ]
    other_kwargs = {
        key: getattr(cube, key, None)
        for key in [
            "attributes",
            "standard_name",
            "long_name",
            "var_name",
            "units",
            "cell_methods",
            "aux_factories",
            "cell_measures_and_dims",
        ]
    }
    cyclic_cube = Cube(
        cy_data,
        dim_coords_and_dims=dim_coords_and_dims,
        aux_coords_and_dims=aux_coords_and_dims,
        **other_kwargs,
    )
    return cyclic_cube


def add_planet_calendar(
    cube,
    time_coord,
    days_in_year,
    days_in_month,
    days_in_day,
    run_start_day=0,
    seasons=("djf", "mam", "jja", "son"),
    planet="planet",
):
    """
    Add an auxiliary time axis with the non-Earth period lengths.

    Parameters
    ----------
    cube: iris.cube.Cube
        Input cube.
    time_coord: iris.coords.Coord or str
        Original time coordinate of the cube.
    days_in_year: int or float
        Number of Earth days in one year on the given planet.
    days_in_month: int or float
        Number of Earth days in one month on the given planet.
    days_in_day: int or float
        Number of Earth days in one day on the given planet.
    run_start_day: int or float, optional
        Earth day of the start of the simulation.
    seasons: tuple, optional
        Sequences of letters corresponding to month names.
    planet: str, optional
        Name of the planet to be used to name the new coordinate.
    """

    def rel_day(coord, value):
        """Get the relative number of the day."""
        start = coord.units.num2date(coord.points[0])
        current = coord.units.num2date(value)
        iday = run_start_day + (current - start).days
        return iday

    def determine_season(coord, value):
        """Determine season from the month number."""
        assert coord.name() == f"{planet}_month"
        for season in seasons:
            if value + 1 in _months_in_season(season):
                return season

    new_coords = {
        "year": lambda c, v: rel_day(c, v) // days_in_year,
        "month": lambda c, v: (rel_day(c, v) % days_in_year) // days_in_month,
        "day": lambda c, v: (rel_day(c, v) % days_in_month) // days_in_day,
        "season": determine_season,
    }
    for key, op in new_coords.items():
        if key == "season":
            coord = f"{planet}_month"
        else:
            coord = time_coord
        add_categorised_coord(cube, f"{planet}_{key}", coord, op)


def area_weights_cube(cube, r_planet=None, normalize=False, model=um):
    """
    Create a cube of area weights for an arbitrary planet.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube with longitude and latitude coordinates
    r_planet: float, optional
        Radius of the planet (m). If not given, an attempt is made
        to get it from the cube metadata.
    normalize: bool, optional
        Normalize areas.
    model: aeolus.model.Model, optional
        Model class with relevant coordinate names.

    Returns
    -------
    iris.cube.Cube
        Cube of area weights with the same metadata as the input cube
    """
    cube = cube.copy()
    ensure_bounds(cube, model=model)
    aw = iris.analysis.cartography.area_weights(cube, normalize=normalize)
    if normalize:
        aw = cube.copy(data=aw)
        aw.rename("normalized_grid_cell_area")
        aw.units = "1"
    else:
        if r_planet is None:
            r = get_planet_radius(cube)
        else:
            r = r_planet
        aw = cube.copy(data=aw * (r / iris.fileformats.pp.EARTH_RADIUS) ** 2)
        aw.rename("grid_cell_area")
        aw.units = "m**2"
    return aw


def check_coords(cubes):
    """Check the cubes coordinates for consistency."""
    # get the names of all coords binned into useful comparison groups
    coord_comparison = iris.analysis._dimensional_metadata_comparison(*cubes)

    bad_coords = coord_comparison["ungroupable_and_dimensioned"]
    if bad_coords:
        raise BadCoordinateError(
            "Coordinates found in one cube that describe "
            "a data dimension which weren't in the other "
            "cube ({}), try removing this coordinate.".format(
                ", ".join(group.name() for group in bad_coords)
            )
        )

    bad_coords = coord_comparison["resamplable"]
    if bad_coords:
        raise BadCoordinateError(
            "Some coordinates are different ({}), "
            "consider resampling.".format(
                ", ".join(group.name() for group in bad_coords)
            )
        )


def coarsen_cube(cube, lon_bins, lat_bins, model=um):
    """
    Block-average cube in longitude and latitude.

    Note: no weighting is applied!

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube with longitude and latitude coordinates.
    lon_bins: array-like
        Longitude bins.
    lat_bins: array-like
        Latitude bins.
    model: aeolus.model.Model, optional
        Model class with relevant coordinate names.

    Returns
    -------
    iris.cube.Cube
    """
    cube_out = cube.copy()
    coord_names = [model.y, model.x]
    cube_out = add_binned_coord(cube_out, model.x, lon_bins)
    cube_out = add_binned_coord(cube_out, model.y, lat_bins)

    # To avoid oversampling on the edges,
    # extract subset within the boundaries of target coords
    for coord, target_points in zip(coord_names, (lat_bins, lon_bins)):
        cube_out = cube_out.extract(
            Constraint(
                **{
                    coord: lambda p: target_points.min()
                    <= p
                    <= target_points.max()
                }
            )
        )

    for coord in coord_names:
        cube_out = cube_out.aggregated_by(
            [f"{coord}_binned"], iris.analysis.MEAN
        )

    for coord, target_points in zip(coord_names, (lat_bins, lon_bins)):
        dim = cube_out.coord_dims(coord)
        units = cube_out.coord(coord).units
        cube_out.remove_coord(coord)
        aux = cube_out.coord(f"{coord}_binned")
        new_points = target_points[aux.points]
        new_coord = DimCoord.from_coord(
            aux.copy(points=new_points, bounds=None)
        )
        cube_out.remove_coord(f"{coord}_binned")
        new_coord.rename(coord)
        new_coord.units = units
        cube_out.add_dim_coord(new_coord, dim)

    return cube_out


def coord_delta_to_cube(cube, coord, normalize=False):
    """
    Convert 1D coord spacings to a cube of the same dim as the given cube.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube containing the coordinate to be broadcast.
    coord: str or iris.coords.Coord
        Coordinate to be broadcast.
    normalize: bool, optional
        Normalize the data.

    Returns
    -------
    iris.cube.Cube
        Cube of broadcast coordinate deltas.
    """
    # Extract the coordinate and its dimension number
    if isinstance(coord, str):
        _coord = cube.coord(coord).copy()
    else:
        _coord = coord.copy()
    dim_map = cube.coord_dims(_coord.name())
    # Ensure the coordinate has bounds
    if not _coord.has_bounds():
        _coord.guess_bounds()

    _data = _coord.bounds[:, 1] - _coord.bounds[:, 0]
    if normalize:
        _data_max = np.nanmax(np.abs(_data))
        _data /= _data_max
        units = "1"
        prefix = "normalized_"
    else:
        units = _coord.units
        prefix = ""
    if len(dim_map) > 0:
        _data = broadcast_to_shape(_data, cube.shape, dim_map)
        dc = [(c.copy(), cube.coord_dims(c)) for c in cube.dim_coords]
        ac = [(c.copy(), cube.coord_dims(c)) for c in cube.aux_coords]
        new_cube = Cube(
            data=_data,
            units=units,
            dim_coords_and_dims=dc,
            aux_coords_and_dims=ac,
        )
    else:
        new_cube = Cube(data=_data, units=units)
    new_cube.rename(f"{prefix}delta_{_coord.name()}")
    return new_cube


def coord_to_cube(cube, coord, broadcast=True):
    """
    Convert coordinate points to a cube with the same dimensions.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube containing the coordinate to be broadcast.
    coord: str or iris.coords.Coord
        Coordinate to be broadcast.
    broadcast: bool, optional
        Broadcast the coordinate points to the shape of the cube.

    Returns
    -------
    iris.cube.Cube
        Cube of broadcast coordinate.
    """
    if isinstance(coord, str):
        _coord = cube.coord(coord)
    else:
        _coord = coord
    kw = {
        "units": _coord.units,
        "long_name": _coord.long_name,
        "standard_name": _coord.standard_name,
        "var_name": _coord.var_name,
    }
    dim_map = cube.coord_dims(_coord.name())
    _data = _coord.points
    if len(dim_map) > 0 and broadcast:
        _data = broadcast_to_shape(_data, cube.shape, dim_map)
        dc = [(c.copy(), cube.coord_dims(c)) for c in cube.dim_coords]
        ac = [(c.copy(), cube.coord_dims(c)) for c in cube.aux_coords]
        new_cube = Cube(
            data=_data, dim_coords_and_dims=dc, aux_coords_and_dims=ac, **kw
        )
    else:
        new_cube = Cube(
            data=_data, dim_coords_and_dims=[(_coord.copy(), 0)], **kw
        )
    return new_cube


def ensure_bounds(cube, coords=("x", "y"), model=um):
    """Auto-generate bounds for cube coordinates."""
    for coord in coords:
        try:
            c_name = getattr(model, coord)
        except (AttributeError, TypeError):
            c_name = coord
        c = cube.coord(c_name)
        if not c.has_bounds():
            if len(c.points) > 1:
                c.guess_bounds()


def get_cube_datetimes(cube, model=um):
    """
    Convert points of a cube's time coordinate to datetimes.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube containing a time coordinate.
    model: aeolus.model.Model, optional
        Model class with relevant coordinate names.

    Returns
    -------
    numpy.ndarray
        Array of datetime-like objects.
    """
    return cube.coord(model.t).units.num2date(cube.coord(model.t).points)


def get_cube_rel_days(cube, model=um):
    """
    Convert points of a cube's time coordinate to relative number of days.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube containing a time coordinate.
    model: aeolus.model.Model, optional
        Model class with relevant coordinate names.

    Returns
    -------
    numpy.ndarray
        Array of relative days.
    """
    dts = get_cube_datetimes(cube, model=model)
    days = ((dts - dts[0]) / timedelta(days=1)).astype(np.float64)
    return days


def get_dim_coord(cube, axis):
    """
    Return a coordinate from a cube based on the axis it represents.

    Uses :py:func:`iris.util.guess_coord_axis` to heuristically match a
    dimensional coordinate with the requested axis.

    Adapted from https://github.com/LSaffin/iris-extensions

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube with the desired coordinate.
    axis: str
        The co-ordinate axis to take from the cube. Must be one of X, Y, Z, T.

    Returns
    -------
    iris.coords.DimCoord
        The dim coordinate matching the requested axis on the given cube.

    Raises
    ------
    ArgumentError: If axis is not one of {X, Y, Z, T}.
    NotFoundError: If the cube does not contain a coord with the given axis.
    """
    _allowed = ["X", "Y", "Z", "T"]
    axis = axis.upper()
    # If the axis supplied is not correct raise an error
    if axis not in _allowed:
        raise ArgumentError(
            f"Axis must be one of {_allowed}, {axis} is given."
        )

    # Loop over dimensional coords in the cube
    for coord in cube.dim_coords:
        # Return the coordinate if it matches the axis
        if axis == guess_coord_axis(coord):
            return coord

    # If no coordinate matches raise an error
    raise NotFoundError(f"Cube has no coordinate for axis {axis}")


def get_xy_coords(cube, model=um):
    """
    Return X and Y coords tuple using names from a given `model` container.

    Parameters
    ----------
    cube: iris.cube.Cube
        Input cube.
    model: aeolus.model.Model, optional
        Model class with relevant coordinate and variable names.
    """
    y_coord = cube.coord(model.y)
    x_coord = cube.coord(model.x)
    return (x_coord, y_coord)


@const_from_attrs(strict=False)
def interp_cube_from_height_to_pressure_levels(
    variable,
    pressure,
    levels,
    p_ref_frac=False,
    const=None,
    interpolator=None,
    model=um,
):
    """
    Interpolate a cube from height to pressure level(s).

    Parameters
    ----------
    variable: iris.cube.Cube
        A cube to interpolate.
        Must have the same dimensions as the pressure cube.
    pressure: iris.cube.Cube
        A cube of pressure.
    levels: array-like
        Sequence of pressure levels
        (same units as the units of pressure cube in `cubelist`).
    p_ref_frac: bool, optional
        If True, `levels` are used as fractions of the surface pressure.
    const: aeolus.const.const.ConstContainer, optional
        Required if p_ref_frac=True.
        If not given, constants are attempted to be retrieved from
        attributes of a cube in the cube list.
    interpolator: callable or None
        The interpolator to use when computing the interpolation.
        See `relevel()` docs for more.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        The variable interpolated to pressure level(s).
    """
    from iris.experimental import stratify

    if p_ref_frac:
        p_ref = const.reference_surface_pressure.copy()
        p_ref.convert_units(pressure.units)
        p_tgt = np.atleast_1d(levels) * float(p_ref.data)
    else:
        p_tgt = np.atleast_1d(levels)
    cube_plev = stratify.relevel(
        variable, pressure, p_tgt, axis=model.z, interpolator=interpolator
    )
    cube_plev.coord(model.pres).attributes = {}
    return squeeze(cube_plev)


@const_from_attrs(strict=False)
def interp_cubelist_from_height_to_pressure_levels(
    cubelist,
    levels,
    p_ref_frac=False,
    const=None,
    interpolator=None,
    model=um,
    include_pressure=False,
):
    """
    Interpolate all cubes to the given set of pressure levels.

    Pressure variable is found using the `model.pres` name and then optionally
    excluded from the output list of interpolated cubes.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        List of cubes, including a cube of pressure.
    levels: array-like
        Sequence of pressure levels
        (same units as the units of pressure cube in `cubelist`).
    p_ref_frac: bool, optional
        If True, `levels` are used as fractions of the surface pressure.
    const: aeolus.const.const.ConstContainer, optional
        Required if p_ref_frac=True.
        If not given, constants are attempted to be retrieved from
        attributes of a cube in the cube list.
    interpolator: callable or None
        The interpolator to use when computing the interpolation.
        See `relevel()` docs for more.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.
    include_pressure: bool, optional
        If True, include the pressure cube in the output.

    Returns
    -------
    iris.cube.CubeList
        List of cubes interpolated to pressure level(s).
    """
    pres = cubelist.extract_cube(model.pres)
    cl_out = CubeList()
    for cube in cubelist:
        if include_pressure or cube != pres:
            cube_plev = interp_cube_from_height_to_pressure_levels(
                cube,
                pres,
                levels,
                p_ref_frac=p_ref_frac,
                const=const,
                interpolator=interpolator,
                model=model,
            )
            cl_out.append(cube_plev)
    return cl_out


def interp_to_cube_time(cube_src, cube_tgt, model=um):
    """
    Interpolate `cube_src` to `cube_tgt` along the time dimension (`model.t`).

    Linear interpolation is used. Forecast period is copied from `cube_tgt`.

    Parameters
    ----------
    cube_src: iris.cube.Cube
        Cube to interpolate.
    cube_tgt: iris.cube.Cube
        Cube with time dimension to interpolate to.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube with the time dimension equal to `cube_tgt`.
    """
    target = [(model.t, cube_tgt.coord(model.t).points)]
    out = cube_src.interpolate(target, iris.analysis.Linear())
    # Replace time coordinates, because interpolation removes bounds.
    for coord in ["t", "fcst_prd", "fcst_ref"]:
        try:
            out.replace_coord(cube_tgt.coord(getattr(model, coord)))
        except (AttributeError, CoNotFound):
            pass
    return out


def isel(obj, coord, idx, skip_not_found=None):
    """
    Slice cubes by an index of a coordinate (index-selector).

    Parameters
    ----------
    obj: iris.cube.Cube or iris.cube.CubeList
        Cube or list of cubes.
    coord: str or iris.coords.Coord
        Coordinate for selection.
    idx: int
        Index along the given coordinate.
    skip_not_found: bool or str, optional
        Skip if coordinate not found. By default it is active when dealing with
        a cube list and inactive if dealing with a single cube.

    Returns
    -------
    iris.cube.Cube or iris.cube.CubeList
        Slice along the coordinate.
    """
    if isinstance(obj, Cube):
        try:
            # _coord = obj.coord(coord)
            # val = _coord.points[idx]
            # try:
            #     val = _coord.units.num2date(val)
            # except ValueError:
            #     pass
            # constr = Constraint(**{_coord.name(): lambda x: x.point == val})
            # out = obj.extract(constr)
            val = obj.coord(coord)[idx]
            out = squeeze(obj.subset(val))
        except CoNotFound as e:
            if skip_not_found:
                out = obj
            else:
                raise e
    elif isinstance(obj, (list, set, tuple)):
        out = []
        for cube in obj:
            out.append(
                isel(
                    cube,
                    coord,
                    idx,
                    skip_not_found=(
                        (skip_not_found is None) or skip_not_found
                    ),
                )
            )
        out = obj.__class__(out)
    return out


def nearest_coord_value(cube, coord_name, val):
    """
    Get the nearest value of a coordinate.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube with the coordinate
    coord_name: str or iris.coords.Coord
        Coordinate where to look the nearest point up
    val: int or float
        The value to find

    Returns
    -------
    int or float
        Element of the coordinate array closest to the given `val`.
    """
    coord = cube.coord(coord_name)
    i = coord.nearest_neighbour_index(val)
    return coord.points[i]


def not_equal_coord_axes(cube1, cube2):
    """Given 2 cubes, return axes of unequal dimensional coordinates."""
    coord_comp = iris.analysis._dimensional_metadata_comparison(cube1, cube2)
    neq_dim_coords = set(coord_comp["not_equal"]).intersection(
        set(coord_comp["dimensioned"])
    )
    dims = []
    for coord_pair in neq_dim_coords:
        for coord in coord_pair:
            dims.append(guess_coord_axis(coord))
    return set(filter(None, dims))


def regrid_3d(cube, target, model=um):
    """
    Regrid a cube in the horizontal and in the vertical.

    Adapted from https://github.com/LSaffin/iris-extensions

    Parameters
    ----------
    cube: iris.cube.Cube
        The cube to be regridded.
    target: iris.cube.Cube
        The cube to regrid to.
    model: aeolus.model.Model, optional
        Model class with relevant coordinate names.

    Returns
    -------
        iris.cube.Cube
    """
    neq_axes = not_equal_coord_axes(cube, target)
    if neq_axes.intersection(["X", "Y"]):
        cube = cube.regrid(target, iris.analysis.Linear())

    # Interpolate in the vertical if needed
    if "Z" in neq_axes:
        vert_coord = model.z
        if vert_coord is None:
            z = get_dim_coord(target, "Z")
        else:
            z = target.coord(vert_coord)
        cube = cube.interpolate([(z.name(), z.points)], iris.analysis.Linear())
        ensure_bounds(cube, coords=[z])

    # Match coordinate information
    # XXX is this needed?
    # newcube = target.copy(data=cube.data)
    # newcube.rename(cube.name())
    # newcube.units = cube.units

    # Put back correct time information
    # for coord in newcube.aux_coords:
    #     if iris.util.guess_coord_axis(coord) == "T":
    #         newcube.remove_coord(coord)
    # for coord in cube.aux_coords:
    #     if iris.util.guess_coord_axis(coord) == "T":
    #         newcube.add_aux_coord(coord)
    return cube


def replace_z_coord(cube, model=um):
    """
    Replace dimensional vertical coordinate.

    Parameters
    ----------
    cube: iris.cube.Cube
        Input cube.
    model: aeolus.model.Model, optional
        Model class with relevant coordinate names.

    Returns
    -------
    iris.cube.Cube
        Copy of the input cube with a new vertical coordinate.
    """
    new_cube = cube.copy()
    new_cube.coord(model.z).bounds = None
    promote_aux_coord_to_dim_coord(new_cube, model.z)
    ensure_bounds(new_cube, coords=[model.z])
    for coord in [model.s, model.lev]:
        # By default, model levels and sigma coordinates are removed.
        try:
            new_cube.remove_coord(coord)
        except CoNotFound:
            pass
    return new_cube


def roll_cube_0_360(cube_in, add_shift=0, model=um):
    """
    Take a cube spanning -180...180 degrees and roll it to 0...360 degrees.

    Works with global model output, and in some cases for regional.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube with longitude and latitude coordinates.
    add_shift: int
        Additional shift of data (is not applied to the longitude coordinate).
    model: aeolus.model.Model, optional
        Model class with a relevant longitude coordinate name.

    Returns
    -------
    iris.cube.Cube

    See Also
    --------
    aeolus.coord.roll_cube_pm180
    """
    cube = cube_in.copy()
    coord_name = model.x  # get the name of the longitude coordinate
    lon = cube.coord(coord_name)
    if (lon.points < 0.0).any():
        add = 180
        cube.data = da.roll(
            cube.core_data(), len(lon.points) // 2 + add_shift, axis=-1
        )
        if lon.has_bounds():
            bounds = lon.bounds + add
        else:
            bounds = None
        new_points = lon.points + add
        cube.replace_coord(lon.copy(points=new_points, bounds=bounds))
    return cube


def roll_cube_pm180(cube_in, add_shift=0, model=um):
    """
    Take a cube spanning 0...360 degrees and roll it to -180...180 degrees.

    Works with global model output, and in some cases for regional.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube with longitude and latitude coordinates.
    add_shift: int
        Additional shift of data (is not applied to the longitude coordinate).
    model: aeolus.model.Model, optional
        Model class with a relevant longitude coordinate name.

    Returns
    -------
    iris.cube.Cube

    See Also
    --------
    aeolus.coord.roll_cube_0_360
    """
    cube = cube_in.copy()
    coord_name = model.x  # get the name of the longitude coordinate
    xcoord = cube.coord(coord_name)
    if (xcoord.points >= 0.0).all():
        assert is_regular(
            xcoord
        ), "Operation is only valid for a regularly spaced coordinate."
        if _is_longitude_global(xcoord.points):
            # Shift data symmetrically only when dealing with global cubes
            cube.data = da.roll(
                cube.core_data(), len(xcoord.points) // 2 + add_shift, axis=-1
            )

        if xcoord.has_bounds():
            bounds = wrap_lons(xcoord.bounds, -180, 360)  # + subtract
            bounds = bounds[bounds[:, 0].argsort(axis=0)]
        else:
            bounds = None
        cube.replace_coord(
            xcoord.copy(
                points=np.sort(wrap_lons(xcoord.points, -180, 360)),
                bounds=bounds,
            )
        )
    else:
        # Nothing to do, the cube is already centered on 0 longitude
        # unless there is something wrong with longitude
        msg = (
            f"Incorrect {coord_name} values: "
            f"from {xcoord.points.min()} to {xcoord.points.max()}"
        )
        assert (
            (xcoord.points >= -180.0) & (xcoord.points <= 180.0)
        ).all(), msg
    return cube


def vertical_cross_section_area(cube2d, r_planet=None):
    """Create a cube of vertical cross-section areas in metres."""
    cube2d = cube2d.copy()
    if r_planet is None:
        r = get_planet_radius(cube2d)
    else:
        r = r_planet
    m_per_deg = (np.pi / 180) * r
    if guess_coord_axis(cube2d.dim_coords[1]) == "X":
        m_per_deg *= np.cos(np.deg2rad(cube2d.coord(axis="Y").points[0]))

    for dim_coord in cube2d.dim_coords:
        if not dim_coord.has_bounds():
            dim_coord.guess_bounds()
    x_bounds = cube2d.coord(cube2d.dim_coords[1]).bounds
    z_bounds = cube2d.coord(cube2d.dim_coords[0]).bounds

    vc_area = cube2d.copy(
        data=(z_bounds[:, 1] - z_bounds[:, 0])[:, None]
        * ((x_bounds[:, 1] - x_bounds[:, 0])[None, :] * m_per_deg)
    )
    vc_area.units = "m**2"
    vc_area.rename("vertical_section_area")
    for dim_coord in vc_area.dim_coords:
        dim_coord.bounds = None
    return vc_area


def volume_weights_cube(cube, r_planet=None, normalize=False, model=um):
    """
    Create a cube of volume weights from a grid of a given cube.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube with longitude, latitude and height coordinates
    r_planet: float, optional
        Radius of the planet (m). If not given, an attempt is made
        to get it from the cube metadata.
    normalize: bool, optional
        Normalize the data.
    model: aeolus.model.Model, optional
        Model class with relevant coordinate names.

    Returns
    -------
    iris.cube.Cube
        Cube of area weights with the same metadata as the input cube
    """
    area_cube = area_weights_cube(
        cube, r_planet=r_planet, normalize=normalize, model=model
    )
    height_deltas = coord_delta_to_cube(cube, model.z, normalize=normalize)
    volume = area_cube * height_deltas
    if normalize:
        volume.rename("normalized_volume_weights")
        volume.convert_units("1")
    else:
        volume.rename("volume_weights")
        volume.convert_units("m**3")
    return volume



"""Generic calculus functions."""

from cf_units import Unit
from iris.analysis import Linear
from iris.analysis.calculus import (
    _coord_cos,
    _curl_differentiate,
    _curl_regrid,
    differentiate,
)
from iris.analysis.maths import add, divide, multiply
from iris.coord_systems import GeogCS, RotatedGeogCS
import numpy as np

from ..const import get_planet_radius
from ..exceptions import NotYetImplementedError
from ..meta import update_metadata
from ..model import um

__all__ = (
    "d_dx",
    "d_dy",
    "d_dz",
    "deriv",
    "div_h",
    "integrate",
)


def d_dx(cube, model=um):
    """Calculate a derivative w.r.t. x-coordinate."""
    return deriv(cube, model.x)


def d_dy(cube, model=um):
    """Calculate a derivative w.r.t. y-coordinate."""
    return deriv(cube, model.y)


def d_dz(cube, model=um):
    """Calculate a derivative w.r.t. z-coordinate."""
    return deriv(cube, model.z)


def deriv(cube, coord):
    """
    Calculate a derivative w.r.t. the given coordinate.

    Uses `iris.analysis.calculus.differentiate` and then
    interpolates the result to the grid points of the original cube.

    Parameters
    ----------
    cube: iris.cube.Cube
        Input cube containing the given coordinate.
    coord: str or iris.coords.Coord
        Coordinate for differentiation.

    Returns
    -------
    iris.cube.Cube
        d(cube)/d(coord).

    See Also
    --------
    aeolus.calc.calculus.d_dx, aeolus.calc.calculus.d_dy
    """
    pnts = cube.coord(coord).points
    diff = differentiate(cube, coord)
    res = diff.interpolate([(coord, pnts)], Linear())
    return res


@update_metadata(name="horizontal_divergence")
def div_h(i_cube, j_cube, r_planet=None, model=um):
    r"""
    Calculate horizontal divergence.

    Note: currently works only in spherical coordinates.

    Parameters
    ----------
    i_cube:
        i-component (e.g. u-wind)
    j_cube:
        j-component (e.g. v-wind)
    r_planet: float, optional
        Radius of the planet (m). If not given, an attempt is made
        to get it from the cube metadata.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        Cube of horizontal divergence.

    Notes
    -----
    Divergence in spherical coordinates is defined as

    .. math::

        \nabla\cdot \vec A = \frac{1}{r cos \theta} (
        \frac{\partial A_\lambda}{\partial \lambda}
        + \frac{\partial}{\partial \theta}
        (A_\theta cos \theta))

    where \lambda is longitude, \theta is latitude.
    """
    x_coord = i_cube.coord(model.x)
    y_coord = i_cube.coord(model.y)

    y_dim = i_cube.coord_dims(y_coord)[0]

    horiz_cs = i_cube.coord_system("CoordSystem")

    # Check for spherical coords
    spherical_coords = isinstance(horiz_cs, (GeogCS, RotatedGeogCS))
    if spherical_coords:
        # Get the radius of the planet
        if r_planet is None:
            r = get_planet_radius(i_cube)
        else:
            r = r_planet
        r_unit = Unit("m")

        lon_coord = x_coord.copy()
        lat_coord = y_coord.copy()
        lon_coord.convert_units("radians")
        lat_coord.convert_units("radians")
        lat_cos_coord = _coord_cos(lat_coord)

        # j-comp: \frac{\partial}{\partial \theta} (\vec A_\theta cos \theta))
        temp = multiply(j_cube, lat_cos_coord, y_dim)
        djcos_dtheta = _curl_differentiate(temp, lat_coord)
        prototype_diff = djcos_dtheta

        # i-comp: \frac{\partial \vec A_\lambda}{\partial \lambda}
        d_i_cube_dlambda = _curl_differentiate(i_cube, lon_coord)
        d_i_cube_dlambda = _curl_regrid(d_i_cube_dlambda, prototype_diff)
        new_lat_coord = d_i_cube_dlambda.coord(model.y)
        new_lat_cos_coord = _coord_cos(new_lat_coord)
        lat_dim = d_i_cube_dlambda.coord_dims(new_lat_coord)[0]

        # Sum and divide
        div = divide(
            add(d_i_cube_dlambda, djcos_dtheta),
            r * new_lat_cos_coord,
            dim=lat_dim,
        )
        div.units /= r_unit
        div = div.regrid(i_cube, Linear())
    else:
        raise NotYetImplementedError(
            "Non-spherical coordinates are not implemented yet."
        )
    return div


def integrate(cube, coord):
    """
    Integrate the cube along a 1D coordinate using the trapezoidal rule.

    Note: `coord` must be one of the dimensional coordinates of the cube.

    Parameters
    ----------
    cube: iris.cube.Cube
        Input cube containing the given coordinate.
    coord: str or iris.coords.Coord
        Coordinate for integration.

    Returns
    -------
    iris.cube.Cube
        Integrated cube.
    """
    # TODO: allow non-dim coordinates
    c = cube.coord(coord)
    others = [
        dc.name()
        for dc in cube.dim_coords
        if cube.coord_dims(dc) != cube.coord_dims(c)
    ]
    dim = cube.coord_dims(c)[0]
    data = np.trapz(cube.data, c.points, axis=dim)
    res = next(cube.slices(others)).copy(data=data)
    res.units = cube.units * c.units
    res.remove_coord(c)
    res.rename(f"integral_of_{cube.name()}_wrt_{c.name()}")
    # ensure_bounds(cube, [c])
    # delta = AuxCoord(c.bounds[:, 1] - c.bounds[:, 0], units=c.units)
    # res = multiply(cube, delta, dim=dim).collapsed(c, iris.analysis.SUM)
    return res



"""Physical constants as collections of scalar `iris` cubes."""

from .const import add_planet_conf_to_cubes, get_planet_radius, init_const

__all__ = ("add_planet_conf_to_cubes", "get_planet_radius", "init_const")
